#include "psublast/libc.h"
#include "sim4.h"
#include "psublast/psublast.h"
#include "sim4b1.h"
#include "scriptlib.h" 
#include "poly.h"


#ifndef __lint
/*@unused@*/
static const char rcsid[] = "$Id: poly.c,v 1.2 1999/05/26 00:12:00 florea Exp $";
#endif

static void remove_polyT_front(struct edit_script_list **,Exon **,uchar *,uchar*);
static void remove_polyA_back(struct edit_script_list **,Exon **,uchar *,uchar*);
static int  adjust_exons_front(Exon **,int,int);
static int  adjust_exons_back(Exon **,int,int);
static void pfree_list(Exon **);

void remove_poly(struct edit_script_list **Script, Exon **Exons, uchar *s1, uchar *s2)
{
     remove_polyT_front(Script, Exons, s1, s2); 
     remove_polyA_back(Script, Exons, s1, s2);

     return;
}

/*
static void adjust_intron_score(struct edit_script_list **Script, Exon **Exons, uchar *seq1, uchar *seq2)
{
    struct edit_script_list *align;

    if (*Script==NULL) return;
    if (*Exons==NULL) 
       { (*Script)->score = 0; return; }
    if (!(*Exons)->to1 && (*Exons)->next_exon && !(*Exons)->next_exon->to1)
       { (*Script)->score = 0; return; }

    align = *Script;
    while (align) {
       Exon *t0=*Exons, *t1;
       int indels = 0, b1, b2, f1, f2;

       f1 = align->offset1; 
       f2 = align->offset2;
       b1 = f1+align->len1-1; 
       b2 = f2+align->len2-1; 
       while (t0 && ((f1!=t0->from1) || (f2!=t0->from2)))
          t0 = t0->next_exon;
       while (t0 && ((b1!=t0->to1) || (b2!=t0->to2))) {
          if (t1=t0->next_exon) 
              indels += t1->from1-t0->to1-1; 
          t0 = t1;
       }
       align->score += indels;
       align = align->next_script;
    }

    return ;
}
*/

static void remove_polyT_front(struct edit_script_list **Sptr, Exon **Exons, uchar *s1, uchar *s2)
{
     const int MIN_NTS = 10; 
     struct edit_script *head, *to_this=NULL;
     struct edit_script_list *align;
     int countT = 0, stop = 0, indels = 0, c, num=0;
     uchar *a, *b;
     
     if ((align=*Sptr)==NULL) return;
     align = *Sptr;
     a = s1 + (*Sptr)->offset1 -1;
     b = s2 + (*Sptr)->offset2 -1;

     head = align->script;
     while (!stop && head) {
        num = head->num;
        switch (head->op_type) {
           case DELETE: a += num; indels += num; num = 0;
                        head = head->next;
/* 
                        while (num && *a=='T')  
                             { a++; countT++; indels++; num--; }
                        if (!num) head = head->next; 
                        else stop = 1; 
*/
                        break; 

           case INSERT: while (num && *b=='T')
                           { b++; countT++; indels++; num--; }
                        if (!num) head = head->next; 
                        else stop = 1;
                        break;

           case SUBSTITUTE:
                        while (num && *a =='T' && *b=='T')
                             { a++; b++; countT++; num--; }
                        if (!num) head = head->next; 
                        else stop = 1;
                        break;

           default: fatal("sim4b1: remove_polyT_front(): Unrecognized optype in the script");  
        }
        to_this = head;
     }

     if (to_this && countT>=MIN_NTS) {
        /* commit to changing the script */
        to_this->num = num; 
        while (to_this && (to_this->op_type!=SUBSTITUTE)) {
           num = to_this->num;
           switch (to_this->op_type) {
               case DELETE: a += num; indels += num;
                            to_this = to_this->next;
                            break;
               case INSERT: b += num; indels += num;
                            to_this = to_this->next;     
                            break;
               case SUBSTITUTE: break;
               default: fatal("sim4b1: remove_polyT_front(): Unrecognized optype in the script");
           }
        }
     }

     if (to_this==NULL) {
         /* removed entire alignment */
         struct edit_script_list *tmp;
    
         if ((align=align->next_script)!=NULL)  {
             c = adjust_exons_front(Exons,align->offset1,align->offset2);
             align->score += c;
         } else {
             pfree_list(Exons); 
         }

         Script_free((*Sptr)->script); tmp = *Sptr;
         *Sptr = (*Sptr)->next_script; free(tmp);
         remove_polyT_front(Sptr, Exons, s1, s2);
         
         return ;
     }

     if (countT>=MIN_NTS) {
         /* commit to removing tail upto next substitution */

         (*Sptr)->len1 = (*Sptr)->len1-(a-s1+1)+(*Sptr)->offset1;
         (*Sptr)->len2 = (*Sptr)->len2-(b-s2+1)+(*Sptr)->offset2;
         (*Sptr)->offset1 = a-s1+1;
         (*Sptr)->offset2 = b-s2+1;
         (*Sptr)->score -= indels;

         head = (*Sptr)->script;
         while (head && (head!=to_this)) {
            (*Sptr)->script = head->next;
            free(head); 
            head = (*Sptr)->script; 
         }
     }

     c = adjust_exons_front(Exons,(*Sptr)->offset1, (*Sptr)->offset2);
     align->score += c;
	    
     return;
}


/* returns the number of intron position erased */
static int adjust_exons_front(Exon **exons, int f1, int f2)
{
     Exon *t0, *t1;
     int indels = 0;

     if (*exons==NULL) return 0;
    
     t0 = (*exons)->next_exon; 
     if ((*exons)->to1 && t0 && t0->to1) return 0;
     while (t0 && ((f2<t0->from2) || (f2>t0->to2))) {

         t1 = t0->next_exon;
         if (t1->to1) indels += t1->from1-t0->to1-1;
         (*exons)->next_exon = t1; free(t0); t0 = t1;
     } 
     if (t0) {
         t0->from1 = f1;
         t0->from2 = f2;
         t0->length = t0->to2-t0->from2+1;
     }
     
     return indels;
}

static void remove_polyA_back(struct edit_script_list **Sptr, Exon **Exons, uchar *s1, uchar *s2)
{
     const int MIN_NTS = 10;
     struct edit_script *head, *to_this=NULL;
     struct edit_script_list *align, *prev;
     int countA = 0, stop = 0, indels = 0, c, num=0;
     uchar *a, *b;
     
     if (*Sptr==NULL) return;
     align = *Sptr; prev = NULL;

     while (align && align->next_script) { 
        prev = align; 
        align = align->next_script; 
     }

     a = s1 + align->offset1 + align->len1 -2;
     b = s2 + align->offset2 + align->len2 -2;
     
     Flip_script(&(align->script));
     head = align->script;

     while (!stop && head) {
        num = head->num;
        switch (head->op_type) {
           case DELETE: a -= num; indels += num; num = 0;
                        head = head->next;
/*                   
                        while (num && *a=='A')
                           { a--; countA++; indels++; num--; }
                        if (!num) head = head->next;
                        else stop = 1;
*/
                        break;
           
           case INSERT: while (num && *b=='A')
                             { b--; countA++; indels++; num--; }
                        if (!num) head = head->next;
                        else stop = 1;
                        break;
           
           case SUBSTITUTE:
                        while (num && *a =='A' && *b=='A') 
                             { a--; b--; countA++; num--; }
                        if (!num) head = head->next;
                        else stop = 1;
                        break;
           
           default: fatal("sim4b1: remove_polyA_back(): Unrecognized optype in the script");
        }  
        to_this = head;
     }  

     if (to_this && countA>=MIN_NTS) {
        /* commit to changing the script */
        to_this->num = num;
        while (to_this && (to_this->op_type!=SUBSTITUTE)) {
           num = to_this->num;
           switch (to_this->op_type) {
               case DELETE: a -= num; indels += num;
                            to_this = to_this->next;
                            break;
               case INSERT: b -= num; indels += num;
                            to_this = to_this->next;
                            break;
               case SUBSTITUTE: break;
               default: fatal("sim4b1: remove_polyT_front(): Unrecognized optype in the script");
           }
        }  
     }
        
     if (to_this==NULL) {
         /* removed entire alignment */
         
         if (prev!=NULL) {
             c = adjust_exons_back(Exons, prev->offset1+prev->len1-1,
                                          prev->offset2+prev->len2-1);
             align->score += c;
         }  else {
             pfree_list(Exons); 
         }
         Script_free(align->script); free(align);
         if (prev) prev->next_script = NULL;
         else *Sptr = NULL;
         remove_polyA_back(Sptr, Exons, s1, s2);

         return ;
     }   
         
     if (countA>=MIN_NTS) {
         /* remove tail: offsets remain the same */ 

         align->len1 = (a-s1+1)-align->offset1+1;
         align->len2 = (b-s2+1)-align->offset2+1;
         align->score -= indels; 
         
         head = align->script;
         while (head && (head!=to_this)) {
            align->script = head->next;  
            free(head);
            head = align->script;
         }
     }   
     Flip_script(&(align->script));
         
     c = adjust_exons_back(Exons,align->offset1+align->len1-1, align->offset2+align->len2-1);
     align->score += c;

     return;
}

static int adjust_exons_back(Exon **exons, int b1, int b2)
{
     Exon *t0, *last, *t1;
     int indels = 0;
 
     if (*exons==NULL) return 0;
     if ((*exons)->to1 && (t0=(*exons)->next_exon) && t0->to1) return 0;

     last = *exons;       
     while (last && last->next_exon) last = last->next_exon;
     flip_list(exons,&last);
 
     t0 = (*exons)->next_exon;
     while (t0 && ((b2<t0->from2) || (b2>t0->to2))) {
         t1 = t0->next_exon;
         if (t1->to1) indels += t0->from1-t1->to1-1;
         if (t0->to1) {
             (*exons)->next_exon = t1; free(t0); t0 = t1;
         } else t0 = t1; 
     }
     if (t0) {
         t0->to1 = b1; t0->to2 = b2;
         t0->length = t0->to2-t0->from2+1;
     }
     flip_list(exons,&last);

     return indels;
}

static void pfree_list(Exon **exons)
{
     Exon *t0;
 
     t0 = (*exons)->next_exon;  
     while (t0 && t0->to1) {
        t0 = t0->next_exon;
        free((*exons)->next_exon);
        (*exons)->next_exon = t0;
     }
     return;
}
